# -*- coding: utf-8 -*-
"""
Payment History service, records modification done on payments
"""
import logging
from sqlalchemy import desc
from endi_payment.interfaces import IPaymentArchiveService
from endi_payment.models import EndiPaymentHistory


class HistoryLogService(object):
    """
    History service that's only logging the actions in a log file.

    To be used in dev mode
    """
    def __init__(self, context, request):
        self.request = request
        self.logger = logging.getLogger("endi_payment")
        self.archive_service = request.find_service(
            IPaymentArchiveService
        )
        self.dbsession = request.dbsession

    def record(self, action, invoice, payment):
        self.logger.debug("Recording an action")
        self.logger.debug("Action : %s" % action)
        self.logger.debug("User : %s" % self.request.identity)
        self.logger.debug("Invoice id : %s" % invoice.id)


class HistoryDBService(HistoryLogService):
    """
    Service logging datas in a database using a specific connexion
    """

    def get_previous_entry_hash(self):
        """
        Produce a hash of the csv representation of the previous history entry

        :param obj dbsession: The current database session
        :rtype: str
        """
        previous_entry = self.dbsession.query(
            EndiPaymentHistory
        ).order_by(desc(EndiPaymentHistory.created_at)).first()

        if previous_entry:
            result = previous_entry.get_hash()
        else:
            result = ""
        print(result)
        return result

    def record(self, action, invoice, payment):
        """
        Record an action on a payment

        1- Record the action
        2- If an archive service exists, creates an archive

        :param str action: The type of action
        :param obj invoice: The related Invoice instance
        :param obj payment: The related payment object
        """
        self.logger.info("Recording an action")
        self.logger.info("Action : %s" % action)
        self.logger.info("User : %s" % self.request.identity)
        self.logger.info("Invoice id : %s" % invoice.id)
        self.logger.info("It's production mode")

        previous_entry_hash = self.get_previous_entry_hash()
        record = EndiPaymentHistory(
            action_type=action,
            payment_id=payment.id,
            mode=payment.mode,
            amount=payment.amount,
            bank_remittance_id=payment.bank_remittance_id,
            date=payment.date,
            invoice_id=invoice.id,
            invoice_pdf_file_hash=invoice.pdf_file_hash,
            bank_cg=payment.bank.compte_cg,
            tva_value=payment.tva.value,
            user_login=self.request.identity.login.login,
            previous_entry_hash=previous_entry_hash,
        )
        self.dbsession.add(record)
        self.dbsession.flush()
        self.archive_service.archive(record)
        self.logger.info("Recorded")



# -*- coding:utf-8 -*-
from pyramid.path import DottedNameResolver

from endi_payment.interfaces import (
    IPaymentRecordHistoryService,
    IPaymentArchiveService,
)


def configure_history_service(config, settings):
    """
    Configure the history service that will be used to log the Payment history
    """
    module_path = "endi_payment.history.HistoryLogService"
    key = "endi_payment.interfaces.IPaymentRecordHistoryService"

    if key in settings:
        module_path = settings[key]

    # si c'est un service de stockage en bdd on doit configurer la session
    # interne au module
    if module_path == "endi_payment.history.HistoryDBService":
        if 'endi_payment_db.url' not in settings:
            raise Exception("endi_payment.dev est à false et aucune url "
                            "endi_payment_db.url n'est fourni")
        else:
            # We will store the endi payment's history
            config.include('.database')

    history_service = DottedNameResolver().resolve(module_path)
    config.register_service_factory(
        history_service, IPaymentRecordHistoryService
    )


def configure_archive_service(config, settings):
    """
    Configure the Archive service that will be used to archive the Payment
    history log
    """
    module_path = "endi_payment.archive.DefaultArchiveService"
    key = "endi_payment.interfaces.IPaymentArchiveService"

    if key in settings:
        module_path = settings[key]

    service = DottedNameResolver().resolve(module_path)

    if hasattr(service, "check_settings"):
        service.check_settings(settings)

    config.register_service_factory(service, IPaymentArchiveService)


def includeme(config):
    settings = config.get_settings()
    configure_history_service(config, settings)
    configure_archive_service(config, settings)



# -*- coding: utf-8 -*-
"""
Endi storage services
"""
import datetime
import hashlib
import os
import logging

from endi_payment.models import EndiPaymentArchiveSeal
from endi_payment.interfaces import IPaymentAsyncArchiveService
from endi.utils.widgets import Link
from endi_payment.views.routes import ENDI_PAYMENT_ARCHIVE_ITEM


class DefaultArchiveService(object):
    ui_panel_name = "endi_payment.local_archive_panel"
    ui_list_item_panel_name = "endi_payment.local_file_archive_list_item_panel"

    def __init__(self, context, request):
        self.request = request
        self.logger = logging.getLogger("endi_payment")
        self.dbsession = request.dbsession

    def archive(self, history):
        self.logger.debug("Archiving history item {}".format(history.id))
        id_key = "No persistent archive system has been used"
        result = EndiPaymentArchiveSeal(
            remote_identification_key=id_key,
            endi_payment_history_id=history.id
        )
        return result

    def is_archived(self, history):
        return False

    def find(self, history):
        """
        Find A EndiPaymentArchiveSeal that matches the given history

        :param obj history: A EndiPaymentHistory instance
        :rtype: class:`endi_payment.models.EndiPaymentArchiveSeal
        """
        query = self.dbsession.query(EndiPaymentArchiveSeal)
        query = query.filter(
            EndiPaymentArchiveSeal.endi_payment_history_id == history.id
        )
        return query.first()

    def get_ui_plugins(self):
        """
        List the panels to use in the history view
        """
        return [self.ui_panel_name]

    def get_ui_list_plugin(self):
        """
        List the panels to use in the history view
        """
        return self.ui_list_item_panel_name

    def stream_list_actions(self, seal):
        if seal:
            yield Link(
                self.request.route_path(
                    ENDI_PAYMENT_ARCHIVE_ITEM, id=seal.id
                ),
                "Télécharger l'archive",
                title="Voir le détail de cette entrée",
                icon="archive"
            )


class FileArchiveService(object):
    """
    FileStorageService simply logs the history in a file
    """
    settings_key = "endi_payment_archive_storage_path"
    ui_panel_name = "endi_payment.local_archive_panel"
    ui_list_item_panel_name = "endi_payment.local_file_archive_list_item_panel"

    def __init__(self, context, request):
        self.logger = logging.getLogger("endi_payment")
        self.request = request
        self.dbsession = request.dbsession
        self.storage_path = self.request.registry.settings[self.settings_key]
        self.filename = self.get_filename()
        self.filepath = os.path.join(self.storage_path, self.filename)

        try:
            self.async_archive_service = request.find_service(
                IPaymentAsyncArchiveService
            )
        except:
            self.logger.exception("No async archive service was configured")
            self.async_archive_service = None

    def get_filename(self, date=None):
        if date is None:
            date = datetime.date.today()
        return "payment_storage_{}_{}.csv".format(
            date.year, date.month
        )

    def get_seal_filepath(self, archive_seal):
        """
        Return the path to the file associated to the given archive_seal

        :param obj archive_seal: The EndiPaymentArchiveSeal instance
        :rtype: str
        """
        filename = self.get_filename(date=archive_seal.created_at)
        return os.path.join(self.storage_path, filename)

    def _get_archive_filecontent(self):
        return open(self.filepath, 'rb').read()

    def _get_id_key(self, filecontent):
        id_key = hashlib.sha1(filecontent).hexdigest()
        return id_key

    def archive(self, history):
        """
        Archive the given history entry

        :returns: A Sha1 sum of the output file content
        :rtype: str
        """
        with open(self.filepath, 'a') as fbuf:
            fbuf.write(history.serialize())

        filecontent = self._get_archive_filecontent()

        if self.async_archive_service is not None:
            self.async_archive_service.archive(
                history, self.filename, filecontent
            )
            # The seal will be produced asynchronously
            result = None
            self.logger.debug("No archive seal was produced locally")
            self.logger.debug("The asynchronous service should handle this")
        else:
            # The seal is produced locally
            id_key = self._get_id_key(filecontent)
            result = EndiPaymentArchiveSeal(
                archive_type='local',
                remote_identification_key=id_key,
                endi_payment_history_id=history.id
            )
            self.dbsession.add(result)
            self.dbsession.flush()
        return result

    def is_archived(self, history):
        """
        Check if the payment history entry has been archived

        :param obj history: The EndiPaymentHistory instance
        :rtype: bool
        """
        query = self.dbsession.query(EndiPaymentArchiveSeal.id)
        query = query.filter(
            EndiPaymentArchiveSeal.endi_payment_history_id == history.id
        )
        result = query.count() > 0
        return result

    def find(self, history):
        """
        Find A EndiPaymentArchiveSeal that matches the given history

        :param obj history: A EndiPaymentHistory instance
        :rtype: class:`endi_payment.models.EndiPaymentArchiveSeal
        """
        query = self.dbsession.query(EndiPaymentArchiveSeal)
        query = query.filter(
            EndiPaymentArchiveSeal.endi_payment_history_id == history.id
        )
        return query.first()

    def get_ui_plugins(self):
        """
        List the panels to use in the history view
        """
        return [self.ui_panel_name]

    def get_ui_list_plugin(self):
        """
        List the panels to use in the history view
        """
        return self.ui_list_item_panel_name

    def stream_list_actions(self, seal):
        if seal:
            if seal.archive_type == 'local':
                yield Link(
                    self.request.route_path(
                        ENDI_PAYMENT_ARCHIVE_ITEM, id=seal.id
                    ),
                    "Télécharger l'archive",
                    title="Voir le détail de cette entrée",
                    icon="archive"
                )
            elif self.async_archive_service and seal.archive_type == \
                self.async_archive_service.archive_type_key:
                for action in self.async_archive_service.stream_list_actions(
                    seal
                ):
                    yield action

    @classmethod
    def check_settings(cls, settings):
        """
        Check the settings contains the endi_payment_archive_storage_path if
        this given service is configured in the ini file

        :raises: KeyError if the key is missing
        :raises: Exception if the directory doesn't exist
        """
        if cls.settings_key not in settings:
            raise KeyError(
                "You should configure {} in your .ini file".format(
                    cls.settings_key
                )
            )

        storage_path = settings[cls.settings_key]
        if not os.path.isdir(storage_path):
            raise Exception(
                "Invalid storage path {}".format(storage_path)
            )



"""5.1.4 Update endi_payment

Revision ID: afcf321a324d
Revises: aaaaaaaaaaaa
Create Date: 2020-03-18 16:40:19.616393

"""

# revision identifiers, used by Alembic.
revision = 'afcf321a324d'
down_revision = 'aaaaaaaaaaaa'

from alembic import op
import sqlalchemy as sa


def update_database_structure():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('endi_payment_archive_seal', sa.Column('archive_type', sa.String(length=20), nullable=True))
    # ### end Alembic commands ###

def migrate_datas():
    from endi_base.models.base import DBSESSION
    session = DBSESSION()
    from alembic.context import get_bind
    conn = get_bind()
    conn.execute('update endi_payment_archive_seal set archive_type="local"')
    from zope.sqlalchemy import mark_changed
    mark_changed(session)

def upgrade():
    update_database_structure()
    migrate_datas()


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('endi_payment_archive_seal', 'archive_type')
    # ### end Alembic commands ###



#-*-coding:utf-8-*-
import logging
import os
import datetime
import random

from endi.scripts.utils import command



def generate_archives_command(arguments, env):
    """
    Generate archive files based on payment database entries
    """
    logger = logging.getLogger(__name__)
    logger.debug("Generating archive files")
    from zope.sqlalchemy import mark_changed
    from endi_base.models.base import DBSESSION
    from endi_payment.archive import FileArchiveService
    from endi_payment.models import (
        EndiPaymentHistory,
        EndiPaymentArchiveSeal,
    )
    storage = env['request'].registry.settings[FileArchiveService.settings_key]
    logger.debug("  + Working with directory {}".format(storage))
    os.system('rm -f {}/*.csv'.format(storage))

    dbsession = DBSESSION()
    seals = dbsession.query(EndiPaymentArchiveSeal).all()
    logger.debug(" Deleting {} seals".format(len(seals)))
    for seal in seals:
        dbsession.delete(seal)

    entries = dbsession.query(EndiPaymentHistory).order_by(
        EndiPaymentHistory.created_at
    ).all()

    logger.debug(" + {} entries".format(len(entries)))
    for index, entry in enumerate(entries):
        if index > 0:
            entry.previous_entry_hash = entries[index - 1].get_hash()
        else:
            entry.previous_entry_hash = ''
        archive_service = FileArchiveService(
            None, env['request']
        )
        archive_service.filename = archive_service.get_filename(
            entry.created_at
        )
        archive_service.filepath = os.path.join(
            archive_service.storage_path, archive_service.filename
        )
        seal = archive_service.archive(entry)
        seal.created_at = entry.created_at + datetime.timedelta(
            microseconds=random.randint(100, 350)
        )
    mark_changed(dbsession)


def entry_point():
    """enDI payment archive generation tool
    Usage:
        endi-payment <config_uri> gen_archive
    """
    try:
        return command(generate_archives_command, entry_point.__doc__)
    finally:
        pass



def includeme(config):
    settings = config.get_settings()
    key = "endi_payment.interfaces.IPaymentRecordHistoryService"
    if key in settings and \
            settings[key] == "endi_payment.history.HistoryDBService":
        config.include(".history")

    key = "endi_payment.interfaces.IPaymentArchiveService"
    if key in settings and \
            settings[key] == "endi_payment.archive.FileArchiveService":
        config.include('.archive')



# -*- coding: utf-8 -*-
from zope.interface import Interface, Attribute


class IPaymentRecordHistoryService(Interface):
    """
    History manipulation tool iinterface for Payment action log
    """

    def record_action(self, action, invoice, payment):
        """
        History manipulation tool for Payment registration
        """
        pass


class IPaymentArchiveService(Interface):
    """
    Archive service is used to archive payment datas and secure its content
    integrity with the help of third-party services
    """

    def archive(self, history):
        """
        Stores a single history item

        :param obj history: A EndiPaymentHistory instance
        :rtype: class:`endi_payment.models.EndiPaymentArchiveSeal
        """
        pass

    def is_archived(self, history):
        """
        Check if the given history item has been archived remotely

        :param obj history: A EndiPaymentHistory instance
        :rtype: bool
        """
        pass

    def find(self, history):
        """
        Find A EndiPaymentArchiveSeal that matches the given history

        :param obj history: A EndiPaymentHistory instance
        :rtype: class:`endi_payment.models.EndiPaymentArchiveSeal
        """

    def get_ui_plugins(self):
        """
        List ui panels that should be used in display

        See pyramid_layout to known what a panel is

        :returns: List of panel names
        :rtype: list
        """

    def get_ui_list_plugin(self):
        """
        Collect the ui panel used to display archive informations

        :returns: The name of a panel
        :rtype: str
        """

    def stream_list_actions(self, history):
        """
        Stream actions available in the history list view

        :returns: List all the available actions
        :rtype: list
        """


class IPaymentAsyncArchiveService(Interface):
    """
    Asynchronous archive service that will remotely backup a given archive file

    This Service is supposed to produce a certification along the file that
    will be uploaded

    The file related information can be stored with the EndiPaymentArchiveSeal
    model
    """
    # Key used in database
    archive_type_key = Attribute("""key used in the database""")

    def archive(self, history, filename, filecontent):
        """
        Archive the given file buffer's content

        :param obj history: A EndiPaymentHistory instance
        :param str filename: The name of the archived file
        :param str filecontent: A read buffer of the content to archive
        :returns: True/False if the async task is launched
        :rtype: bool
        """
        pass

    def get_ui_plugins(self):
        """
        List ui panels that should be used in display

        See pyramid_layout to known what a panel is

        :returns: List of panel names
        :rtype: list
        """

    def stream_list_actions(self, history):
        """
        Stream actions available in the history list view

        :returns: List all the available actions
        :rtype: list
        """



"""
Panel associated to the FileArchiveService
"""
import logging
import io
import csv
from endi.export.utils import write_file_to_request

from endi_payment.archive import FileArchiveService
from endi_payment.models import EndiPaymentArchiveSeal
from .routes import ENDI_PAYMENT_ARCHIVE_ITEM


TYPE_LABELS = {
    'local': "Locale",
    "edocgroup": "Distante (Edoc Group)",
}

logger = logging.getLogger(__name__)


def split_archive_on_history_id(filepath, history_id):
    """
    Renvoie le contenu d'un fichier d'archive jusqu'à la ligne contenant
    l'entrée de journal history_id

    :rtype: io.BytesIO instance
    """
    lines = []
    with open(filepath, 'r') as fbuf:
        reader = csv.reader(fbuf)
        for line in reader:
            lines.append(line)
            if line[0] == str(history_id):
                break

    str_result = io.StringIO()
    writer = csv.writer(str_result)
    writer.writerows(lines)
    result = io.BytesIO(str_result.getvalue().encode('utf-8'))
    return result


def archive_file_view(request):
    """
    Stream an archive file content associated to a given archive_seal
    """
    seal_id = request.matchdict['id']
    archive_seal = request.dbsession.query(
        EndiPaymentArchiveSeal
    ).filter_by(id=seal_id).one()

    archive_service = FileArchiveService(None, request)
    filepath = archive_service.get_seal_filepath(archive_seal)
    data_buffer = split_archive_on_history_id(
        filepath, archive_seal.endi_payment_history_id
    )
    write_file_to_request(
        request,
        "{}.csv".format(archive_seal.remote_identification_key),
        data_buffer,
        "text/csv"
    )
    return request.response


def local_file_archive_panel(context, request):
    """
    Collect datas to display local archive informations

    :rtype: dict
    """
    logger.debug("Archive panel")
    archive_seal = request.dbsession.query(
        EndiPaymentArchiveSeal
    ).filter_by(endi_payment_history_id=context.id).first()

    result = {
        'archive_seal': archive_seal,
    }

    if archive_seal:
        result['archive_type_label'] = TYPE_LABELS.get(
            archive_seal.archive_type,
            archive_seal.archive_type,
        )
        archive_service = FileArchiveService(context, request)
        result['filename'] = archive_service.filename

        if archive_seal.archive_type == 'local':
            result['file_link'] = request.route_path(
                ENDI_PAYMENT_ARCHIVE_ITEM,
                id=archive_seal.id
            )
        else:
            async_service = archive_service.async_archive_service
            if async_service:
                result['panels'] = async_service.get_ui_plugins()

    logger.debug(result)
    return result


def local_file_archive_list_item_panel(context, request):
    """
    Collect data to display archive related informations in the history list

    :rtype: dict
    """
    if context:
        result = {
            'archive_seal': context,
            'archive_type_label': TYPE_LABELS.get(
                context.archive_type, context.archive_type
            )
        }
    else:
        result = {}

    return result


def includeme(config):
    config.add_view(
        archive_file_view,
        route_name=ENDI_PAYMENT_ARCHIVE_ITEM,
        permission="admin_treasury",
    )

    config.add_panel(
        local_file_archive_panel,
        'endi_payment.local_archive_panel',
        renderer='endi_payment:views/templates/local_archive_panel.mako'
    )
    config.add_panel(
        local_file_archive_list_item_panel,
        'endi_payment.local_file_archive_list_item_panel',
        renderer='endi_payment:views/templates/\
local_archive_list_item_panel.mako'
    )



"""5.1.4 Update endi_payment

Revision ID: aaaaaaaaaaaa
Revises: None
Create Date: 2020-03-18 16:40:19.616393

"""

# revision identifiers, used by Alembic.
revision = 'aaaaaaaaaaaa'
down_revision = None

from alembic import op
import sqlalchemy as sa


def update_database_structure():
    # ### commands auto generated by Alembic - please adjust! ###
    # ### end Alembic commands ###
    pass

def migrate_datas():
    pass

def upgrade():
    update_database_structure()
    migrate_datas()


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###



# -*- coding: utf-8 -*-
import datetime
import csv
import hashlib
import io

from sqlalchemy import (
    Column,
    Integer,
    BigInteger,
    String,
    DateTime,
    Text,
    ForeignKey,
)
from sqlalchemy.orm import relationship
from sqla_inspect.ascii import force_encoding
from endi_base.models.base import default_table_args
from endi_payment.database import ModelBase


class EndiPaymentHistory(ModelBase):
    __tablename__ = "endi_payment_history"
    __table_args__ = default_table_args
    id = Column(Integer, primary_key=True)
    created_at = Column(DateTime(), default=datetime.datetime.now)
    # ADD/UPDATE/DELETE
    action_type = Column(String(6))
    # Original payment instance id
    payment_id = Column(Integer, nullable=False)
    # Payment mode
    mode = Column(String(50))
    # The amount of the payment (* 10 000)
    amount = Column(BigInteger(), nullable=False)
    # The string identifying the bank remittance id
    bank_remittance_id = Column(String(255))
    # Date the payment has been received
    date = Column(DateTime(), nullable=False)
    # The invoice database item's id
    invoice_id = Column(Integer, nullable=False)
    # The invoice PDF sha1 hash that can be used to check pdf files
    invoice_pdf_file_hash = Column(String(255), nullable=False)
    # The bank CG account
    bank_cg = Column(String(120), default="")
    # The TVA associated to the recorded payment
    tva_value = Column(Integer, nullable=False)
    # The login of the user responsible for this modification
    user_login = Column(String(64), nullable=False)
    # The previous entry's md5sum
    previous_entry_hash = Column(String(255), nullable=False)
    endi_payment_archive_seal = relationship(
        "EndiPaymentArchiveSeal",
        primaryjoin="EndiPaymentHistory.id=="
        "EndiPaymentArchiveSeal.endi_payment_history_id",
        uselist=False,
        back_populates="endi_payment_history"
    )

    def as_csv(self):
        """
        Returns a csv representation of this entry

        :rtype: str
        """
        data_list = [
            self.id,
            self.created_at.isoformat(),
            self.action_type,
            self.payment_id,
            self.mode,
            self.amount,
            self.bank_remittance_id,
            self.date.isoformat(),
            self.invoice_id,
            self.invoice_pdf_file_hash,
            self.bank_cg,
            self.tva_value,
            self.user_login,
            self.previous_entry_hash,
        ]
        line = io.StringIO()

        #data_list = [force_encoding(i, 'utf-8') for i in data_list]

        writer = csv.writer(line)
        writer.writerow(data_list)
        return line.getvalue()

    def serialize(self):
        """
        Serialize the current entry

        :rtype: str
        """
        return self.as_csv()

    def as_binary(self):
        """
        Builds a binary representation of the current model's csv string

        :rtype: bytes
        """
        return self.as_csv().encode('utf-8')

    def get_hash(self):
        return hashlib.sha1(self.as_binary()).hexdigest()


class EndiPaymentArchiveSeal(ModelBase):
    __tablename__ = "endi_payment_archive_seal"
    __table_args__ = default_table_args
    id = Column(Integer, primary_key=True)
    created_at = Column(DateTime, default=datetime.datetime.now)
    archive_type = Column(String(20), default='local')
    remote_identification_key = Column(Text(), nullable=False)
    endi_payment_history_id = Column(ForeignKey('endi_payment_history.id'))

    endi_payment_history = relationship(
        EndiPaymentHistory,
        primaryjoin="EndiPaymentHistory.id=="
        "EndiPaymentArchiveSeal.endi_payment_history_id",
        back_populates="endi_payment_archive_seal"
    )



import datetime
import colander
import logging
from sqlalchemy import extract

from endi.views import BaseListView
from endi_payment.models import (
    EndiPaymentHistory,
    EndiPaymentArchiveSeal,
)
from endi.utils.widgets import Link, ButtonLink
from endi.models.task import Invoice, Payment

from endi_payment.interfaces import IPaymentArchiveService
from .forms import get_list_schema
from .routes import (
    ENDI_PAYMENT_HISTORY,
    ENDI_PAYMENT_HISTORY_ITEM,
    ENDI_PAYMENT_HISTORY_ITEM_ARCHIVE,
)


logger = logging.getLogger(__name__)


ACTION_LABELS = {
    'ADD': 'Ajout', 'UPDATE': 'Modification', 'DELETE': 'Suppression'
}


class HistoryListView(BaseListView):
    title = "Journal des modifications sur les encaissements"
    add_template_vars = (
        'title',
        'action_labels',
        'stream_actions',
        "archive_item_panel",
    )
    schema = get_list_schema()
    default_sort = "created_at"
    default_direction = "desc"
    sort_columns = {
        'created_at': 'created_at',
        'user_login': 'user_login',
        'action_type': 'action_type',
        'tva_value': 'tva_value',
        'amount': 'amount',
    }

    null_values = (colander.null, None, '')
    action_labels = ACTION_LABELS

    def __init__(self, *args, **kwargs):
        BaseListView.__init__(self, *args, **kwargs)
        self.archive_service = self.request.find_service(
            IPaymentArchiveService
        )
        self.archive_item_panel = self.archive_service.get_ui_list_plugin()

    def query(self):
        return self.dbsession.query(EndiPaymentHistory).outerjoin(
            EndiPaymentArchiveSeal
        )

    def filter_created_at(self, query, appstruct):
        period = appstruct.get('period', {})
        if period.get('start') not in self.null_values:
            logger.debug("  + Filtering by date : %s" % period)
            start = period.get('start')
            end = period.get('end')
            if end not in self.null_values:
                end = datetime.date.today()
            query = query.filter(
                EndiPaymentHistory.created_at.between(start, end)
            )
        else:
            year = appstruct.get('year')
            if year not in self.null_values:
                query = query.filter(
                    extract('year', EndiPaymentHistory.created_at) == year
                )
        return query

    def filter_amount(self, query, appstruct):
        amounts = appstruct.get('amount', {})
        if amounts.get('start') not in self.null_values:
            logger.debug("Filtering by amount : %s" % amounts)
            start = amounts.get('start')
            end = amounts.get('end')
            if end in self.null_values:
                query = query.filter(EndiPaymentHistory.amount >= start)
            else:
                query = query.filter(
                    EndiPaymentHistory.amount.between(start, end)
                )
        return query

    def filter_action_type(self, query, appstruct):
        action_type = appstruct.get('action_type')
        if action_type not in self.null_values:
            query = query.filter(EndiPaymentHistory.action_type == action_type)
        return query

    def filter_invoice_id(self, query, appstruct):
        invoice_id = appstruct.get('invoice_id')
        if invoice_id:
            logger.debug("Filtering by invoice_id {}".format(invoice_id))
            query = query.filter(EndiPaymentHistory.invoice_id == invoice_id)
        return query

    def filter_invoice_number(self, query, appstruct):
        invoice_number = appstruct.get('official_number')
        if invoice_number:
            invoice_id = self.dbsession.query(Invoice.id).filter(
                Invoice.official_number == invoice_number
            ).scalar()

            logger.debug(
                "Filtering by invoice_id {} (number {})".format(
                    invoice_id, invoice_number
                )
            )
            query = query.filter(EndiPaymentHistory.invoice_id == invoice_id)
        return query

    def stream_actions(self, item):
        yield Link(
            self.request.route_path(ENDI_PAYMENT_HISTORY_ITEM, id=item.id),
            "Voir",
            title="Voir le détail de cette entrée",
            icon="arrow-right"
        )
        seal = item.endi_payment_archive_seal
        for action in self.archive_service.stream_list_actions(seal):
            yield action


def history_item_view(request):
    """
    EndiPaymentHistory item view
    """
    logger.debug("endi_payment.EndiPaymentHistory view")
    history_id = request.matchdict['id']
    history = request.dbsession.query(EndiPaymentHistory).get(history_id)
    request.actionmenu.add(
        ButtonLink('Journal des modifications', ENDI_PAYMENT_HISTORY)
    )
    archive_service = request.find_service(IPaymentArchiveService)
    archive_seal = archive_service.find(history)

    panels = archive_service.get_ui_plugins()

    return dict(
        title="Vue détaillée d'une entrée du journal de modification des "
        "encaissements",
        history=history,
        payment=Payment.get(history.payment_id),
        invoice=Invoice.get(history.invoice_id),
        action_labels=ACTION_LABELS,
        archive=archive_seal,
        panels=panels
    )


def includeme(config):
    config.include('.routes')

    config.add_view(
        HistoryListView,
        route_name=ENDI_PAYMENT_HISTORY,
        permission='admin_treasury',
        renderer="endi_payment:views/templates/histories.mako"
    )
    config.add_view(
        history_item_view,
        route_name=ENDI_PAYMENT_HISTORY_ITEM,
        permission="admin_treasury",
        renderer="endi_payment:views/templates/history.mako"
    )

    config.add_admin_menu(
        parent="accounting",
        permission="admin_treasury",
        label="Journaux des encaissements",
        href=ENDI_PAYMENT_HISTORY,
    )



import colander
import deform

from endi.forms import (
    range_validator,
    deferred_default_year,
    year_filter_node,
)
from endi.forms.widgets import CleanMappingWidget
from endi.forms.tasks.lists import PeriodSchema, AmountRangeSchema
from endi.forms.lists import BaseListsSchema
from endi.models.task.invoice import get_invoice_years


def get_list_schema():
    """
    Build a form schema used to list history entries
    """
    schema = BaseListsSchema().clone()
    del schema['search']

    node = colander.SchemaNode(
        colander.String(),
        name='official_number',
        title="Numéro de facture",
        missing=colander.drop
    )
    schema.insert(0, node)
    node = colander.SchemaNode(
        colander.Integer(),
        name='invoice_id',
        title="",
        widget=deform.widget.HiddenWidget(),
        missing=colander.drop
    )
    schema.insert(0, node)

    node = PeriodSchema(
        name="period",
        title="",
        validator=colander.Function(
            range_validator,
            msg="La date de début doit précéder la date de début"
        ),
        widget=CleanMappingWidget(),
        missing=colander.drop,
    ).clone()
    node['start'].title = "Enregistré entre le"
    schema.insert(0, node)
    node = AmountRangeSchema(
        name='amount',
        title="",
        validator=colander.Function(
            range_validator,
            msg="Le montant de départ doit être inférieur ou égale "
            "à celui de la fin"
        ),
        widget=CleanMappingWidget(),
        missing=colander.drop,
    ).clone()
    node['start'].title = "Montant entre"
    schema.insert(0, node)

    node = colander.SchemaNode(
        colander.String(),
        name='action_type',
        title="Type d'action",
        widget=deform.widget.SelectWidget(
            values=(
                ('', 'Tous'),
                ('ADD', 'Ajout'),
                ('UPDATE', 'Modification'),
                ('DELETE', 'Suppression'),
            )
        ),
        missing=colander.drop,
    )
    schema.insert(0, node)

    node = year_filter_node(
        name='year',
        title="Année",
        query_func=get_invoice_years,
        default=deferred_default_year,
        default_val=('', 'Tous'),
        missing=colander.drop,
    )
    schema.insert(0, node)


    return schema



import os

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'README.rst')) as f:
    README = f.read()

with open(os.path.join(here, 'requirements.txt')) as f:
    requires = f.read()

with open(os.path.join(here, 'CURRENT_VERSION')) as f:
    current_version = f.read().splitlines()[0].strip()


tests_require = [
    'WebTest >= 1.3.1',  # py3 compat
    'pytest',  # includes virtualenv
    'pytest-cov',
]


entry_points = {
    "console_scripts": [
        "endi-payment = endi_payment.scripts:entry_point"
    ]
}


setup(name='endi_payment',
      version=current_version,
      description='endi_payment',
      long_description=README,
      long_description_content_type="text/x-rst",
      license='GPLv3',
      classifiers=[
          "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
          "Programming Language :: Python",
          "Framework :: Pyramid",
          "Topic :: Internet :: WWW/HTTP",
          "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
      ],
      author='Majerti',
      author_email='equipe@majerti.fr',
      url='https://framagit.org/endi/endi_payment',
      keywords='web wsgi bfg pylons pyramid endi',
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      extras_require={
          'testing': tests_require,
      },
      install_requires=requires,
      entry_points=entry_points,
      )



# -*- coding: utf-8 -*-
"""
Specific session stuff used to store the payment history logs
"""
import logging

from sqlalchemy import engine_from_config
from sqlalchemy.ext.declarative import declarative_base

logger = logging.getLogger(__name__)
ModelBase = declarative_base()


def includeme(config):
    """
    Pyramid Include's mechanism
    Setup the library specific session
    """
    logger.debug("Setting up database configuration for endi_payment")
    settings = config.get_settings()
    from endi_base.models.base import DBBASE, DBSESSION

    if 'endi_payment_db.url' in settings:
        # On a une connexion particulière pour l'édition des journaux
        prefix = 'endi_payment_db.'
        endi_payment_engine = engine_from_config(settings, prefix=prefix)
        main_bind = DBSESSION.bind

        # Pour éviter un warning
        # https://bitbucket.org/zzzeek/sqlalchemy_old/issues/3977/
        DBSESSION.remove()

        DBSESSION.configure(
            binds={ModelBase: endi_payment_engine, DBBASE: main_bind},
        )

    else:
        # On utilise l'engine sqlalchemy par défaut (celui d'endi)
        # Pas besoin de spécifier le bind qui est le même qu'avant
        endi_payment_engine = DBSESSION.bind

    # Pour que le create_all fonctionne on doit importer les modèles
    from endi_payment.models import EndiPaymentHistory  # NOQA
    ModelBase.metadata.bind = endi_payment_engine
    ModelBase.metadata.create_all(endi_payment_engine)
    return True



# -*- coding: utf-8 -*-
from alembic import context
import traceback
import transaction

from endi.alembic.exceptions import MigrationError, RollbackError
from endi_base.models.base import DBSESSION
from endi.models import DBBASE


def run_migrations_online():
    from endi_payment.models import EndiPaymentHistory
    from endi_payment.database import ModelBase
    bind = DBSESSION.get_bind(EndiPaymentHistory)
    if bind is None:
        raise ValueError(
"\nYou must do enDI migrations using the 'endi-migrate' script"
"\nand not through 'alembic' directly."
            )

    transaction.begin()
    connection = DBSESSION.connection(mapper=EndiPaymentHistory)

    context.configure(
        connection=connection,
        target_metadata=ModelBase.metadata,
        compare_type=True,
    )

    try:
        context.run_migrations()
    except Exception as migration_e:
        traceback.print_exc()
        try:
            transaction.abort()
        except Exception as rollback_e:
            traceback.print_exc()
            raise RollbackError(rollback_e)
        else:
            raise MigrationError(migration_e)
    else:
        transaction.commit()
    finally:
        #connection.close()
        pass

run_migrations_online()



import os

ENDI_PAYMENT_HISTORY = "/endi_payment/history"
ENDI_PAYMENT_HISTORY_ITEM = os.path.join(ENDI_PAYMENT_HISTORY, "{id}")
ENDI_PAYMENT_HISTORY_ITEM_ARCHIVE = os.path.join(
    ENDI_PAYMENT_HISTORY_ITEM, "archive"
)

ENDI_PAYMENT_ARCHIVE = "/endi_payment/archives"
ENDI_PAYMENT_ARCHIVE_ITEM = os.path.join(ENDI_PAYMENT_ARCHIVE, "{id}")


def includeme(config):
    config.add_route(ENDI_PAYMENT_HISTORY, ENDI_PAYMENT_HISTORY)
    config.add_route(ENDI_PAYMENT_HISTORY_ITEM, ENDI_PAYMENT_HISTORY_ITEM)
    config.add_route(
        ENDI_PAYMENT_HISTORY_ITEM_ARCHIVE,
        ENDI_PAYMENT_HISTORY_ITEM_ARCHIVE,
    )
    config.add_route(ENDI_PAYMENT_ARCHIVE_ITEM, ENDI_PAYMENT_ARCHIVE_ITEM)
